# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.


from signals import SignalDispatcher


class TeamType(SignalDispatcher):

    def __init__(self, team):
        SignalDispatcher.__init__(self)

    """
    This class represents the base class for Teamgeist Team objects
    """

    def GetName(self):
        """
        Get the name of the team. This is equivalent to the MUC's name
        """
        raise NotImplementedError()

    def SetName(self, name):
        """
        Set the name of the team. This is equivalent to the MUC's name
        """
        raise NotImplementedError()

    def GetSubject(self):
        """
        Get the name of the team. This is equivalent to the MUC's subject
        """
        raise NotImplementedError()

    def SetSubject(self, subject):
        """
        Set the name of the team. This is equivalent to the MUC's subject
        """
        raise NotImplementedError()

    def GetMembers(self):
        """
        Returns the list of all the members currently in the team
        """
        raise NotImplementedError()

    def InviteMember(self, user):
        """
        Invite a user to become a member of the team.
        If the current user does not have enough priviledges to invite that
        user, then this will send a request to the owner of the team to
        invite that user
        """
        raise NotImplementedError()

    def KickMember(self, user):
        """
        Kick a user from the team.
        If the current user does not have enough priviledges to kick that
        user, then this will send a request to the owner of the team to
        kick that user
        """
        raise NotImplementedError()

    def JoinTeam(self):
        """
        Join the team.
        This should set the team as joinable in the database and attempt to join
        it if the account is available and connected
        """
        raise NotImplementedError()

    def LeaveTeam(self):
        """
        Leave the team.
        This should leave the MUC and make the team non-joinable in the database
        without deleting it and without deleting the associated topics
        """
        raise NotImplementedError()

    def SuggestTopicToMember(self, topic, member):
        """
        Suggests to a member of the team a topic for him to share
        """
        raise NotImplementedError()

    def SuggestTopicToTeam(self, topic):
        """
        Suggests to all members of the team a topic to be shared
        """
        raise NotImplementedError()

    def RequestDocument(self, member, document):
        """
        Request a document from a member of the team
        """
        raise NotImplementedError()

    def ShareEvent(self, event):
        """
        Share the event with the members of the team
        """
        raise NotImplementedError()

    def IsCurrentlyParticipating(self):
        """ Returns whether or not the user is currently participating
        in the team
        """
        raise NotImplementedError()

    @staticmethod
    def GetDescription():
        """
        Return a description of this team type
        """
        raise NotImplementedError()

    __signals__ = {
        "NameChanged":
            "Signal sent when the name of the team is changed",
        "SubjectChanged":
            "Signal sent when the subject of the team is changed",
        "MemberJoined":
            "Signal sent when a member joins the team",
        "MemberLeft":
            "Signal sent when a member has left the team",
        "UserRequestsInvitation":
            "Signal sent when a user requests you to invite him to the team",
        "MemberRequestsKick":
            """Signal sent when a member requests you to kick someone from
            the team
            """,
        "TopicSuggestionReceived":
            "Signal sent when a team member suggests a topic for you to share",
        "DocumentRequestReceived":
            "Signal sent when a team member requests a document from you",
        "ParticipationChanged":
            """Signal sent when Teamgeist participation in a team changes
            (joining or leaving the team)
            """,
        "EventReceived":
            "Signal sent when an event is received from a team member",
        }
    # TODO: add signals for when the list of authorized members changes?
    # needed to know when someone gets kicked... ?
