# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.



from bases import *
from telepathy._generated.Client import Client
from telepathy._generated.Client_Handler import ClientHandler
from telepathy._generated.Client_Approver import ClientApprover
from telepathy.server.properties import DBusProperties
import telepathy.interfaces
import telepathy.constants
import telepathy

from utils import do_reply
from signals import SignalDispatcher

# TODO: how do we get this to get destroyed automatically ?
# TODO: I need those damn singletons!

class DBusChannel(BaseChannel,
                  ReadyableObject,
                  SignalDispatcher):

    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseChannel.__init__(self, bus_name, object_path, bus)
        SignalDispatcher.__init__(self)
        ReadyableObject.__init__(self)

        self.self_handle = None
        self.participants = {}
        self.bus_name_to_handle = {}
        self._got_dbus_names = False
        self.parameters = None

    def BaseReady(self):
        self[telepathy.CHANNEL_INTERFACE_GROUP].get_property(
            "SelfHandle",
            reply_handler=self._got_self_handle_cb,
            error_handler=self._got_self_handle_error)

        self[telepathy.CHANNEL_TYPE_DBUS_TUBE].get_property(
            'DBusNames',
            reply_handler=self._got_dbus_names_cb,
            error_handler=self._got_dbus_names_error)

        self[telepathy.CHANNEL_INTERFACE_TUBE].get_property(
            'Parameters',
            reply_handler=self._got_parameters_cb,
            error_handler=self._got_parameters_error)

    def _got_self_handle_cb(self, self_handle):
        self.self_handle = int(self_handle)
        print "self handle found : ", self_handle
        if self._got_dbus_names and self.parameters:
            self.ready()

    def _got_dbus_names_cb(self, names):
        self.DBusNamesChanged(names, ())
        self._got_dbus_names = True
        if self.self_handle and self.parameters:
            self.ready()

    def _got_parameters_cb(self, parameters):
        self.parameters = parameters
        print "Got parameters : ", self.parameters
        if self.self_handle and self._got_dbus_names:
            self.ready()

    def _got_dbus_names_error(self, error):
        logging.basicConfig()
        logger.error('Get DBusNames property failed: %s', error)
        self.ready()

    def _got_self_handle_error(self, error):
        logging.basicConfig()
        logger.error('GetSelfHandle failed: %s', error)
        self.ready()

    def _got_parameters_error(self, error):
        logging.basicConfig()
        logger.error('Get Parameters property failed: %s', error)
        self.ready()

    def DBusNamesChanged(self, added, removed):
        for handle, bus_name in added.items():
            self.participants[int(handle)] = bus_name
            self.bus_name_to_handle[bus_name] = int(handle)
            self.emit_signal("ParticipantAdded", handle, bus_name)

        for handle in removed:
            bus_name = self.participants.pop(int(handle), None)
            self.bus_name_to_handle.pop(bus_name, None)
            self.emit_signal("ParticipantRemoved", handle, bus_name)

        print "Total participants (%r) : %d" % (self, len(self.participants))


    def Closed(self):
        # TODO: listen to NameOwnerChanged to also send the Closed signal
        # when gabble crashes
        print "Channel closed"
        self.emit_signal("Closed")
        self.disconnect_all_signals()

class Channel(BaseChannel,
              SignalDispatcher):
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseChannel.__init__(self, bus_name, object_path, bus)
        SignalDispatcher.__init__(self)

    def Closed(self):
        # TODO: listen to NameOwnerChanged to also send the Closed signal
        # when gabble crashes
        self.emit_signal("Closed")

class Connection(BaseConnection,
                 ReadyableObject,
                 SignalDispatcher):
    def __init__(self, bus_name = None, object_path = None, bus = None,
                 error_handler=default_error_handler):
        BaseConnection.__init__(self, bus_name, object_path, bus, error_handler)
        SignalDispatcher.__init__(self)
        ReadyableObject.__init__(self)
        self._channels = {}
        self._handle = None

    def BaseReady(self):
        self._got_channels = False
        def error():
            self.ready()
        def got_self_handle(self_handle):
            self._handle = self_handle
            if self._got_channels:
                self.ready()
        def got_channels(channels):
            for path, props in channels:
                self._channels[path] = props
            self._got_channels = True
            if self._handle:
                self.ready()

        self.get_property("SelfHandle",
                          reply_handler=got_self_handle,
                          error_handler=error)


        self[telepathy.interfaces.CONNECTION_INTERFACE_REQUESTS].get_property(
            "Channels",
            reply_handler=got_channels,
            error_handler=error)

    @property
    def handle(self):
        return self._handle

    @property
    def channels(self):
        return self._channels

    def NewChannels(self, channels):
        for path, props in channels:
            print "New channel : ", path
            self._channels[path] = props
            self.emit_signal("NewChannel", path)

    def ChannelClosed (self, channel):
        print "Channel closed : ", channel
        try:
            # We may get a ChannelClosed for a channel we didn't have
            del self._channels[channel]
            self.emit_signal("ChannelClosed", channel)
        except:
            pass

    def get_stored_members(self, reply_handler, error_handler):
        def _error(error):
            do_reply(error_handler, error)
        def _get_stored_cb(yours, path, properties):
            def _get_members_cb(handles):
                do_reply(reply_handler, handles)

            channel = self.get_related_object(Channel, path)
            channel[telepathy.interfaces.CHANNEL_INTERFACE_GROUP]. \
                get_property('Members',
                             reply_handler=_get_members_cb,
                             error_handler=_error)


        self[telepathy.interfaces.CONNECTION_INTERFACE_REQUESTS]. \
            EnsureChannel({
                telepathy.CHANNEL + '.ChannelType':
                    telepathy.CHANNEL_TYPE_CONTACT_LIST,
                telepathy.CHANNEL + '.TargetHandleType':
                    telepathy.HANDLE_TYPE_LIST,
                telepathy.CHANNEL + '.TargetID':
                    "stored",
                },
                          reply_handler=_get_stored_cb,
                          error_handler=_error)

    def request_handles(self, ids, reply_handler, error_handler):
        self.RequestHandles(telepathy.HANDLE_TYPE_CONTACT,
                            ids,
                            reply_handler=lambda handles: do_reply(reply_handler,
                                                                   handles),
                            error_handler=lambda error: do_reply(error_handler,
                                                                 error))

    def release_handles(self, handles, reply_handler, error_handler):
        self.ReleaseHandles(telepathy.HANDLE_TYPE_CONTACT,
                            handles,
                            reply_handler=lambda: do_reply(reply_handler),
                            error_handler=lambda error: do_reply(error_handler,
                                                                 error))

    def inspect_handles(self, handles, reply_handler, error_handler):
        self.InspectHandles(telepathy.HANDLE_TYPE_CONTACT,
                            handles,
                            reply_handler=lambda uids: do_reply(reply_handler,
                                                               uids),
                            error_handler=lambda error: do_reply(error_handler,
                                                                 error))

    def get_aliases(self, handles, reply_handler, error_handler):
        self[telepathy.interfaces.CONNECTION_INTERFACE_ALIASING].GetAliases(
            handles,
            reply_handler=lambda aliases: do_reply(reply_handler,
                                                   aliases),
            error_handler=lambda error: do_reply(error_handler,
                                                 error))

    def get_presences(self, handles, reply_handler, error_handler):
        self[telepathy.interfaces.CONNECTION_INTERFACE_SIMPLE_PRESENCE]. \
            GetPresences(handles,
                         reply_handler=lambda presences: do_reply(reply_handler,
                                                                  presences),
                         error_handler=lambda error: do_reply(error_handler,
                                                              error))


class Account(BaseAccount,
              ReadyableObject,
              SignalDispatcher):
    def __init__(self, bus_name = None, object_path = None, bus = None):
        BaseAccount.__init__(self, bus_name, object_path, bus)
        ReadyableObject.__init__(self)
        SignalDispatcher.__init__(self)
        self._display_name = None
        self._normalized_name = None
        self._responses = 0

    def BaseReady(self):
        self.get_property('DisplayName',
                          reply_handler=self._get_display_name_cb,
                          error_handler=self._get_property_error)
        self.get_property('NormalizedName',
                          reply_handler=self._get_normalized_name_cb,
                          error_handler=self._get_property_error)
        self.get_property("Connection",
                          reply_handler=self._get_connection_cb,
                          error_handler=self._get_property_error)

    def _get_display_name_cb(self, display_name):
        self._display_name = display_name
        self._responses += 1
        if self._responses == 3:
            self.ready()

    def _get_normalized_name_cb(self, normalized_name):
        if normalized_name != "":
            self._normalized_name = normalized_name
        self._responses += 1
        if self._responses == 3:
            self.ready()

    def _get_connection_cb(self, connection):
        self._connection = None
        if connection is not None and \
                connection != "" and connection != "/":
            self._connection = Connection(object_path=connection)
            self.emit_signal("ConnectionChanged")
        self._responses += 1
        if self._responses == 3:
            self.ready()

    def _get_property_error(self, error):
        self._responses += 1
        if self._responses == 3:
            self.ready()
        raise error

    @property
    def connection(self):
        return self._connection

    @property
    def display_name(self):
        return self._display_name

    @property
    def normalized_name(self):
        return self._normalized_name

    def __repr__(self):
        if self._display_name:
            return self._display_name
        elif self._normalized_name:
            return self._normalized_name
        else:
            return self.object_path

    def __str__(self):
        return self.__repr__()

    # Signals!
    def Removed(self):
        pass

    def AccountPropertyChanged(self, properties):
        if properties.has_key('Connection'):
            if properties['Connection'] == "" or properties['Connection'] == "/":
                self._connection = None
            else:
                self._connection = Connection(
                    object_path=properties['Connection'])
            self.emit_signal("ConnectionChanged")

        if properties.has_key('DisplayName'):
            self._display_name = properties['DisplayName']

        if properties.has_key('NormalizedName'):
            self._display_name = properties['NormalizedName']

class AccountManager(BaseAccountManager,
                     ReadyableObject,
                     SignalDispatcher):
    def __init__(self):
        BaseAccountManager.__init__(self)
        ReadyableObject.__init__(self)
        SignalDispatcher.__init__(self)
        self._accounts = {}
        self.get_property("ValidAccounts",
                          reply_handler= self._get_valid_accounts_cb,
                          error_handler= self._get_valid_accounts_error)

    def _get_valid_accounts_cb(self, accounts):
        # TODO : TeamType should report which connection managers it can support
        # so we can have it more extensible and add support for non-gabble CMs
        for account_obj in accounts:
            if account_obj. \
                    startswith('/org/freedesktop/Telepathy/Account/gabble/'):
                account = self.get_related_object(Account, account_obj)
                self._accounts[account_obj] = account
                self.emit_signal("NewAccount", account_obj, account)
        self.ready()

    def _get_valid_accounts_error(self, error):
        raise error

    @property
    def accounts(self):
        return self._accounts

    def AccountRemoved(self, object_path):
        if self._accounts.has_key(object_path):
            self.emit_signal("AccountDeleted",
                             object_path, self._accounts[object_path])
            del self._accounts[object_path]

    def AccountValidityChanged	(self, account_obj, valid):
        if valid and not self._accounts.has_key(account_obj) and account_obj. \
                startswith('/org/freedesktop/Telepathy/Account/gabble/'):
            account = self.get_related_object(Account, account_obj)
            self._accounts[account_obj] = account
            self.emit_signal("NewAccount", account_obj, account)

