# -.- coding: utf-8 -.-

# Teamgeist
#
# Copyright (C) 2009 Collabora Limited
# *  Contact: Youness Alaoui
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
# * Contributors:
# *   Youness Alaoui, Collabora Ltd.
# *   Robert McQueen, Collabora Ltd.
# *   Seif Lotfy


from team_type import TeamType
from signals import SignalDispatcher
from utils import do_reply
from bases import ReadyableObject, AsyncMethodBuilder
from team import Team
from helpers import RemoteEvent, Member, Contacts
import dbus
import dbus.service
import dbus.connection

import telepathy.interfaces
import telepathy.constants
import telepathy
import gobject
import random

class Xmpp_muc(TeamType):

    def __init__(self, team):
        TeamType.__init__(self, team)
        self.team = team
        self.account = self.team.account
        self.muc_id = self.team.uid
        self.teamgeist = self.team.teamgeist
        self.tp_client = self.teamgeist._tp_client
        self.participation = Team.PARTICIPATION_NOT_REQUESTED
        self.tp_client.connect("HandleTube", self._got_tube)
        self._text = None
        self._incoming_tubes = []
        self._tube = None
        self._bus_addr = None
        self._bus_conn = None
        self._dbus_obj = None
        self._objs = {}
        self._timeout_id = None
        self._requesting = False

    def join(self):
        def got_error(error):
            # TODO: different participation depending on error?
            self.ParticipationChanged(Team.PARTICIPATION_FAILED)
        if self._text == None:
            def got_reply(channel):
                if self._text:
                    print "Already have a text channel"
                else:
                    self._text = channel
                    self._text.connect("Closed", self._text_closed)
                self.join()
            # TODO: check if self.account._obj is not None
            self.tp_client.request_muc(self.account._obj, self.team.uid,
                                       got_reply, got_error)
        else:
            self._maybe_request_tube()

    def _maybe_request_tube(self):
        self._timeout_id = None
        print "maybe request a tube?"
        if self.participation == Team.PARTICIPATION_NOT_REQUESTED:
            print "we were not asked to participate"
        if self._tube:
            print "not requesting a tube, we already got one"
        elif len(self._incoming_tubes) > 0:
            self._try_accepting_tube()
        else:
            if self._requesting:
                print "A request has already been sent, waiting"
                return False
            print "Requesting a tube"
            def got_reply():
                self._requesting = False
            def got_error(error):
                # TODO: different participation depending on error?
                self._requesting = False
                self.ParticipationChanged(Team.PARTICIPATION_FAILED)
            self._requesting = True
            self.tp_client.request_tube(self.account._obj, self.team.uid,
                                        got_reply, got_error)
        return False

    def _error(self, error):
        print "Got error ", error

    def _try_accepting_tube(self):
        print "Trying to accept a tube"
        if self.participation == Team.PARTICIPATION_NOT_REQUESTED:
            print "we were not asked to participate"
            return False
        elif self._tube:
            print "we already have a tube"
        else:
            if self._text is None:
                if len(self._incoming_tubes) == 0:
                    self.join()
            else:
                if len(self._incoming_tubes) > 0:
                    tube = self._incoming_tubes[0]
                    def accept_tube():
                        print "Accepting tube"
                        tube[telepathy.CHANNEL_TYPE_DBUS_TUBE].Accept(
                            telepathy.constants.SOCKET_ACCESS_CONTROL_LOCALHOST,
                            reply_handler=self._got_tube_address,
                            error_handler=self._error)
                    self._tube = tube
                    tube.call_when_ready(accept_tube)
                else:
                    print "no incoming tubes found"
                    if self.participation != Team.PARTICIPATION_NOT_REQUESTED:
                        self.join()

    def _reset(self):
        self._text = None
        def got_error(tube, error):
            print "Error closing tube %r : %r" % (tube, error)
        def got_closed(tube):
            print "tube %r got closed" % (tube)
        if self._tube not in self._incoming_tubes:
            self._tube.Close(reply_handler=(got_closed, self._tube),
                             error_handler=(got_error, self._tube))
        for tube in self._incoming_tubes:
            tube.Close(reply_handler=(got_closed, tube),
                       error_handler=(got_error, tube))

    def _text_closed(self, text):
        text.disconnect("Closed", self._text_closed)
        if self._text == text:
            self._reset()
        else:
            print "Unknown text channel closed"

    def _tube_closed(self, tube):
        tube.disconnect("Closed", self._tube_closed)
        unknown = False
        if tube in self._incoming_tubes:
            tube.disconnect("ParticipantRemoved",
                            self._tube_participant_left)
            self._incoming_tubes.remove(tube)
        else:
            unknown = True

        if self._tube == tube:
            print "our accepted tube got closed"
            self._tube = None
            self._bus_conn = None
            self._bus_addr = None
            self._dbus_obj = None
            self._objs = {}
            if self.participation == Team.PARTICIPATION_JOINED:
                self.ParticipationChanged(Team.PARTICIPATION_JOINING)
        elif unknown:
            print "unknown tube %r got closed" % tube

        if self._tube is None:
            self._try_accepting_tube()

    def _close_tube(self, tube):
        def got_error(error):
            print "Error closing tube %r : %r" % (tube, error)
        def got_closed():
            print "tube %r got closed" % (tube)
        tube.Close(reply_handler=got_closed,
                   error_handler=got_error)

    def _got_tube(self, emitter, account, muc_id, tube, props):
        if self.account._uid != account or self.muc_id != muc_id:
            return

        tube.connect("Closed", self._tube_closed)
        if props[telepathy.CHANNEL + ".Requested"] == True:
            self._got_tube_channel(tube, props)
        else:
            self._got_tube_offer(tube, props)

    def _got_tube_offer(self, tube, props):
        print "Received a tube offer %r" % tube
        def got_ready():
            if tube not in self._incoming_tubes:
                self._incoming_tubes.append(tube)
                def participant_left(tube, handle, bus_name):
                    if len(tube.participants) == 0:
                        print "This tube is now empty, let's close it"
                        self._close_tube(tube)
                tube.connect("ParticipantRemoved", participant_left)
            if self._tube:
                print "Already have a tube %r" % (self._tube)
                print self._tube.participants
                if len(self._tube.participants) == 1 and \
                        self._tube.self_handle in self._tube.participants:
                    if len(tube.participants) == 1:
                        print "we're both alone in our tubes, let's tiebreak it"
                        if self._tube.parameters['tiebreaker'] < \
                                tube.parameters['tiebreaker']:
                            self._close_tube(self._tube)
                        else:
                            print "my tube is better than theirs, nyanyanya"
                    else:
                        print "I'm alone in this tube, let's close it"
                        self._close_tube(self._tube)

            else:
                self._try_accepting_tube()
        tube.call_when_ready(got_ready)

    def _got_tube_channel(self, tube, props):
        print "got tube channel %r" % tube
        def got_ready():
            if self._tube:
                print "Already have a tube, closing the new one"
                def got_error(error):
                    print "Error closing my tube", error
                def got_closed():
                    print "my tube got closed"
                tube.Close(reply_handler=got_closed,
                           error_handler=got_error)
            else:
                self._tube = tube
                print "Offering tube %r" % tube
                tiebreaker = random.randint(0, 65535)
                tube.parameters = dbus.Dictionary({
                        "tiebreaker": dbus.UInt32(tiebreaker)
                        }, signature="sv")
                tube[telepathy.CHANNEL_TYPE_DBUS_TUBE].Offer(
                    tube.parameters,
                    telepathy.constants.SOCKET_ACCESS_CONTROL_LOCALHOST,
                    reply_handler=self._got_tube_address,
                    error_handler=self._error)
        tube.call_when_ready(got_ready)

    def _got_tube_address(self, bus_addr):
        print "got tube address ", bus_addr
        self._bus_addr = bus_addr
        self._bus_conn = TubeConnection(self, bus_addr)
        self._dbus_obj = TeamgeistTube(self, self._bus_conn)
        for participant in self._tube.participants.keys():
            if participant != self._tube.self_handle:
                self._new_tube_client(participant,
                                      self._tube.participants[participant])

        self._tube.connect("ParticipantRemoved", self._tube_participant_left)
        self._tube.connect("ParticipantAdded", self._tube_participant_joined)
        self.ParticipationChanged(Team.PARTICIPATION_JOINED)

    def _new_tube_client(self, participant, bus_name):
        obj = TeamgeistTubeClient(self._bus_conn, bus_name)
        self._objs[participant] = obj
        member = Member(self.account._obj.connection, participant)
        def got_error(error):
            print "error catching up on events for %s : " % participant, error
            # TODO: should we not listen or ignore the lost events *forever*
            obj.connect("NewEvent", (self._tube_event_received, participant))
        def got_uid(uid):
            def got_events(events):
                print "Got Events for user %s : %s" % (uid, events)
                for event in events:
                    self._tube_event_received(participant, obj, event)
                obj.connect("NewEvent", (self._tube_event_received, participant))

            event_id = self.teamgeist._dbaccessor. \
                GetLastAddedEventForUser(self.team, uid)
            obj.GetEvents(event_id,
                          reply_handler=got_events,
                          error_handler=got_error)

        self.emit_signal("MemberJoined", member)
        member.GetUID(got_uid, got_error)

    def _tube_participant_joined(self, tube, handle, bus_name):
        print "A participant joined tube %r" % tube
        if handle not in self._objs.keys():
            if handle != self._tube.self_handle:
                self._new_tube_client(handle, bus_name)

    def _tube_participant_left(self, tube, handle, bus_name):
        print "A participant left tube %r" % tube
        if handle in self._objs.keys():
            member = Member(self.account._obj.connection, handle)
            self.emit_signal("MemberLeft", member)
            del self._objs[handle]

    def _tube_event_received(self, participant, emitter, dbus_event):
        event = (dbus_event[0], dbus_event[1], "".join(map(str, dbus_event[2])))
        print "Received new event from %d : %r" % (participant, event)
        # TODO: use some kind of singleton per handle
        member = Member(self.account._obj.connection, participant)
        def got_uid(uid):
            remote_event = RemoteEvent.CreateFromStruct(self.team, uid, event)
            self.emit_signal("EventReceived", member, remote_event)
        def got_error(error):
            remote_event = RemoteEvent.CreateFromStruct(self.team,
                                                      participant, event)
            self.emit_signal("EventReceived", member, remote_event)

        member.GetUID(got_uid, got_error)

    @property
    def tube(self):
        return self._tube

    def GetName(self):
        """
        Get the name of the team. This is equivalent to the MUC's name
        """
        # TODO: need text channel...
        # channel telepathy property 10 : 'subject'
        raise NotImplementedError()

    def SetName(self, name):
        """
        Set the name of the team. This is equivalent to the MUC's name
        """
        # channel telepathy property 10 : 'subject'
        raise NotImplementedError()

    def GetSubject(self):
        """
        Get the name of the team. This is equivalent to the MUC's subject
        """
        raise NotImplementedError()

    def SetSubject(self, subject):
        """
        Set the name of the team. This is equivalent to the MUC's subject
        """
        raise NotImplementedError()

    def GetMembers(self, reply_handler, error_handler):
        """
        Returns the list of all the members currently in the team
        """
        do_reply(reply_handler,
                 Contacts(self.account._obj.connection, self._objs.keys()))

    def InviteMember(self, user):
        """
        Invite a user to become a member of the team.
        If the current user does not have enough priviledges to invite that
        user, then this will send a request to the owner of the team to
        invite that user
        """
        raise NotImplementedError()

    def KickMember(self, user):
        """
        Kick a user from the team.
        If the current user does not have enough priviledges to kick that
        user, then this will send a request to the owner of the team to
        kick that user
        """
        raise NotImplementedError()

    def JoinTeam(self, reply_handler, error_handler):
        """
        Join the team.
        This should set the team as joinable in the database and attempt to join
        it if the account is available and connected
        """
        if self.participation == Team.PARTICIPATION_JOINED:
            do_reply(reply_handler, self.participation)
        elif self.participation == Team.PARTICIPATION_JOINING:
            do_reply(reply_handler, self.participation)
        elif self.team.account._obj is None:
            self.ParticipationChanged(Team.PARTICIPATION_ACCOUNT_NOT_EXIST)
            do_reply(error_handler, self.participation)
        else:
            self.ParticipationChanged(Team.PARTICIPATION_JOINING)
            self.join()
            do_reply(reply_handler, self.participation)

    def LeaveTeam(self, reply_handler, error_handler):
        """
        Leave the team.
        This should leave the MUC and make the team non-joinable in the database
        without deleting it and without deleting the associated topics
        """
        # TODO: use the reply/error handlers
        self.ParticipationChanged(Team.PARTICIPATION_NOT_REQUESTED)
        if self.tube:
            self._reset()

    def SuggestTopicToMember(self, topic, member):
        """
        Suggests to a member of the team a topic for him to share
        """
        raise NotImplementedError()

    def SuggestTopicToTeam(self, topic):
        """
        Suggests to all members of the team a topic to be shared
        """
        raise NotImplementedError()

    def RequestDocument(self, member, document):
        """
        Request a document from a member of the team
        """
        raise NotImplementedError()

    def ShareEvent(self, event):
        """
        Share the event with the members of the team
        """
        if self.participation == Team.PARTICIPATION_JOINED:
            self._dbus_obj.NewEvent(event.struct)


    def GetCurrentParticipation(self):
        """ Returns whether or not the user is currently participating
        in the team
        """
        return self.participation

    def ParticipationChanged(self, participation):
        if participation != self.participation:
            self.participation = participation
            self.emit_signal("ParticipationChanged", participation)

    @staticmethod
    def GetDescription():
        """
        Return a description of this team type
        """
        return "XMPP Multi User Chats"


    __signals__ = {
        "NameChanged": # TODO
            "Signal sent when the name of the team is changed",
        "SubjectChanged": # TODO
            "Signal sent when the subject of the team is changed",
        "MemberJoined": # DONE
            "Signal sent when a member joins the team",
        "MemberLeft": # DONE
            "Signal sent when a member has left the team",
        "UserRequestsInvitation":
            "Signal sent when a user requests you to invite him to the team",
        "MemberRequestsKick":
            """Signal sent when a member requests you to kick someone from
            the team
            """,
        "TopicSuggestionReceived":
            "Signal sent when a team member suggests a topic for you to share",
        "DocumentRequestReceived":
            "Signal sent when a team member requests a document from you",
        "ParticipationChanged": # DONE
            """Signal sent when Teamgeist participation in a team changes
            (joining or leaving the team)
            """,
        "EventReceived": # DONE
            "Signal sent when an event is received from a team member",
        }
    # TODO: add signals for when the list of authorized members changes?
    # needed to know when someone gets kicked... ?


class TubeConnection(dbus.connection.Connection):

    def __new__(cls, muc_team, address):
        self = super(TubeConnection, cls).__new__(cls, address)

        self._muc_team = muc_team
        self._tube = self._muc_team.tube
        self.set_unique_name(self._tube.participants[self._tube.self_handle])
        return self

class TeamgeistTube(dbus.service.Object):
    """
    This class defines a Dbus object to be exported over a Tube
    """
    OBJECT_PATH = "/org/gnome/zeitgeist/teamgeist"
    TEAMGEIST_INTERFACE = "org.gnome.zeitgeist.Teamgeist.TUBE"

    def __init__(self, xmpp_tube, connection):
        dbus.service.Object.__init__(self, connection, self.OBJECT_PATH)
        self._xmpp_tube = xmpp_tube

    @dbus.service.signal(dbus_interface=TEAMGEIST_INTERFACE,
                         signature='(asaasay)')
    def NewEvent(self, event):
        """
        This signal is sent whenever the current member wants to share an
        event with the rest of the team
        """
        pass


    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='u', out_signature='a(asaasay)')
    def GetEvents(self, since_id):
        """
        This method is used when a team member joins the team and wants to request
        to catch up with the events he might have mised.
        The since_id parameter specifies the id of the event he last received.
        If since_id is 0, then all events should be returned
        """
        print "events requested since id : ", since_id
        ids = self._xmpp_tube.teamgeist._dbaccessor. \
            GetLocalEventIdsSinceEvent(self._xmpp_tube.team, since_id)
        print "Got new event ids: ", ids
        events = self._xmpp_tube.teamgeist._dbaccessor. \
            GetEvents(self._xmpp_tube.team, None, ids)
        print "Got events : ", events
        def to_struct(event):
            return event.struct
        return map(to_struct, events)

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='s', out_signature='')
    def RequestInvite(self, user):
        """
        This method should be called by a member who does not have enough
        priviledges in the team when he wants to invite someone to the team.
        It should be called on the dbus object shared by the member with the
        highest priviledges.
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='s', out_signature='')
    def RequestKick(self, user):
        """
        This method should be called by a member who does not have enough
        priviledges in the team when he wants to kick someone from the team.
        It should be called on the dbus object shared by the member with the
        highest priviledges.
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='a(suv)', out_signature='')
    def SuggestTopic(self, topic):
        """
        This method is used to suggest a topic to be shared to
        """
        # LATER
        raise NotImplementedError()

    @dbus.service.method(dbus_interface=TEAMGEIST_INTERFACE,
                         in_signature='s', out_signature='')
    def RequestDocument(self, document):
        """
        This method will be called when a user requests a document to be shared
        with him.
        If the document is shareable with this user or team, then a file transfer
        should start in another channel with the requesting user. If not, the
        user will be prompter to share the document and may or may not accept.
        """
        # LATER
        raise NotImplementedError()




class TeamgeistTubeClient(dbus.proxies.ProxyObject,
                          SignalDispatcher):
    """
    A proxy client to the TeamgeistTube object
    """

    def __init__(self, connection, bus_name,
                 object_path = TeamgeistTube.OBJECT_PATH):
        SignalDispatcher.__init__(self)
        dbus.proxies.ProxyObject.__init__(self, connection, bus_name, object_path)
        self.__iface = AsyncMethodBuilder(self, TeamgeistTube.TEAMGEIST_INTERFACE,
                                          ['GetEvents',
                                           'RequestInvite',
                                           'RequestKick',
                                           'SuggestTopic',
                                           'RequestDocument'])
        self.__signal_match = self.__iface.connect_to_signal("NewEvent",
                                                             self.NewEvent)

    def NewEvent(self, event):
        self.emit_signal("NewEvent", event)


# TODO: make this better.. make it as part of the class init, and maybe use
# metaclass and type() to do it...
Team.RegisterType("xmpp_muc", Xmpp_muc)
